%
% This file implements the method described in:
%
% "Robust Region Detection via Consensus Segmentation of Deformable Shapes"
% E. Rodola, S. Rota Bulo, and D. Cremers
% Computer Graphics Forum, 33(5), 2014
%
% If you use this code, please cite the paper above.
% 
% Emanuele Rodola (c) 2014
%
function [clusters,elapsed,iters] = hierarchical_consensus(M, keys, params)

k = params.consensus_max_clusters;

[unique_keys, w] = hash_keys(keys', M.S_tri); % generates base segments
n_subseg = length(w);

% 'w' contains the areas of sub-segments

fprintf('Base segments: %d\n', n_subseg);

if n_subseg>=1000
    warning('hierarchical_consensus(): Too many base segments. Consider reducing the size of the initial ensemble to avoid memory issues!');
end
assert(n_subseg<2000);

samples_per_feature = n_subseg;
samples = repmat((1:n_subseg)',1,n_subseg);

rows = kron(1:n_subseg,ones(1,samples_per_feature))';
cols = samples(:);

weights = w(rows).*w(cols);
wdiag = w.*w;

% delete diagonal elements
offdiag = (rows~=cols);
rows = rows(offdiag);
cols = cols(offdiag);
weights = weights(offdiag);

% W_{ij}: weight for the pair of sub-segments (i,j)
W = full(sparse(rows,cols,weights)) + diag(wdiag);

coassoc = ( kron(unique_keys,ones(n_subseg,1)) == repmat(unique_keys,n_subseg,1) );

tic;

vals = mean(coassoc, 2);
vals = vals(offdiag);

% Cbar_{ij}: fraction of times sub-segments i and j appeared in the same cluster
Cbar = full(sparse(rows,cols,vals)) + eye(n_subseg);

fprintf('Running consensus clustering...');
[Y,iters,~] = mex_PPC_l2(Cbar .* W, W, k, params.consensus_error, params.consensus_max_iters);
fprintf(' done.\n');

elapsed = toc;

% get the final assignment to clusters as m.a.p. and relabel the final 
% clusters from 1 to l<=k

%figure(), imagesc(Y)
[~,subseg_clusters] = max(Y);

subseg_clusters = 1 + subseg_clusters - min(subseg_clusters);

assert(max(subseg_clusters) <= k);

newc = 0;
for i=2:k
    ind = (subseg_clusters==i);
    if newc==0 && ~any(ind)
        %fprintf('Label %d missing\n',i);
        newc = i;
    end
    if newc>0 && any(ind)
        %fprintf('Relabeling %d as %d\n',i,newc);
        subseg_clusters(ind) = newc;
        newc = newc + 1;
    end
end

% convert back from subsegment-wise clusters to triangle-wise clusters
[~,c] = ismember(keys,unique_keys,'rows');
clusters = subseg_clusters(c);

end
